<?php

/*  bodystructure.php:  routines to parse the imap bodystructure response 
 *
    Copyright (C) 2002-2005  Hastymail Development group

    This file is part of Hastymail.

    Hastymail is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Hastymail is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Hastymail; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

* $Id: bodystructure.php,v 1.40 2005/07/26 17:29:34 sailfrog Exp $
*/

/**
 *  simple function to start the parse process
 *
 *  @param          $array       array      bodystructure response in atoms
 *  @return                      array      multi-dimen structure array
 *
 */
function hm_parse_bodystructure($array) {
    while (isset($array[0]) && strtoupper($array[0]) != 'BODYSTRUCTURE') {
        array_shift($array);
    }
    array_shift($array);
    array_pop($array);
    $struct = hm_parse_top_level($array, 1);
    return $struct;
}

/**
 *  function to get the paren "depth" of a subsection of the bodystructure response
 *
 *  @param          $array      array       current section of the bodystructure array
 *  @return                     int         depth level
 *
 */
function hm_get_depth($array) {
    $depth = 0;
    foreach ($array as $val) {
        if ($val == '(') {
            $depth++;
        }
        else {
            break;
        }
    }
    return $depth;
}

/**
 *  increase the part number correctly (1 -> 2, and 1.2 -> 1.3)
 *
 *  @param          $part       string      current part number
 *  @return                     string      new part number
 *
 */
function hm_advance_part_number($part) {
    if (!strstr($part, '.')) {
        $part++;
    }
    else {
        $parts = explode('.', $part);
        $parts[(count($parts) - 1)]++;
        $part = implode('.', $parts);
    }
    return $part;
}

/**
 *  our own array merge function
 *
 *  @param          $array      array       array to add to
 *  @param          $array_2    array       array to add to the first array
 *  @return                     array       combined array
 *
 */
function hm_array_merge($array, $array_2) {
    foreach ($array_2 as $key => $val) {
        $array[$key] = $val;
    }
    return $array;
}

/**
 *  nuts and bolts of the parser. recursively called to break up message parts
 *
 *  @param          $array      array       whats left of the bodystructure response
 *  @param          $part       string      current part number
 *  @return                     array       a part of the multi-dimen structure array
 *
 */
function hm_parse_top_level($array, $part) {
    $struct = array();
    $depth = hm_get_depth($array);
    if ($depth > 1 || (isset($array[$depth]) && 
        strtolower($array[$depth]) == 'message')) {
        if ($depth > 1) {
            $split_array = hm_break_one_level($array);
        }
        else {
            $split_array[0] = $array;
        }
        foreach ($split_array as $array) {
            $depth = hm_get_depth($array);
            if ($depth > 1) {
                $struct = hm_array_merge($struct, hm_parse_top_level($array, $part.'.1'));
                $part = hm_advance_part_number($part);
            }
            else {
                if (isset($array[$depth])) {
                    switch (strtolower($array[$depth])) {
                        case 'binary':
                        case 'text':
                        case 'application':
                        case 'image':
                        case 'audio':
                        case 'video':
                            $struct[$part] = hm_parse_single($array, $part);
                            $part = hm_advance_part_number($part);
                            break;
                        case 'message':
                            switch (strtolower($array[$depth + 1])) {
                                case 'delivery-status':
                                case 'external-body':
                                case 'disposition-notification':
                                case 'rfc822-headers':
                                    $struct[$part] = hm_parse_single($array, $part);
                                    break;
                                default:
                                    $struct[$part] = hm_parse_multi($array, $part);
                                    break;
                            }
                            $part = hm_advance_part_number($part);
                            break;
                        default:
                            if (isset($array[$depth + 1])) {
                                switch (strtolower($array[$depth + 1])) {
                                    case 'unknown':
                                    case 'octet-stream':
                                        $struct[$part] = hm_parse_single($array, $part);
                                        $part = hm_advance_part_number($part);
                                        break;
                                }
                            }
                            break;
                    }
                }
            }
        }               
    }
    else {
        $struct[$part] = hm_parse_single($array, $part);
        $part = hm_advance_part_number($part);
    }
    return $struct;
}

/**
 *  break up a multipart rfc822 message
 *
 *  @param          $array      array       whats left of the bodystructure response
 *  @param          $part       string      current part number
 *  @return                     array       a part of the multi-dimen structure array
 *
*/
function hm_parse_multi($array, $part) {
    $params = array();
    $struct = array();
    $addys = array();
    array_shift($array);
    hm_assign_value($struct, $array, 'type', false, true);
    hm_assign_value($struct, $array, 'subtype', false, true);
    if ($array[0] == '(') {
        array_shift($array);
        while ($array[0] != ')') {
            $params[] = $array[0];
            array_shift($array);
        }
        array_shift($array);
        if (!empty($params)) {
            $count = count($params);
            for ($i=0;$i<$count;$i+=2) {
                if (isset($params[$i]) && isset($params[$i + 1])) {
                    $struct[strtolower($params[$i])] = $params[$i + 1];
                }
            }
        }
    }
    else {
        array_shift($array);
    }
    $data = array('id', 'description', 'encoding', 'rawsize');
    foreach ($data as $key) {
        if (isset($array[0])) {
            hm_assign_value($struct, $array, $key);
        }
    }
    $struct['size'] = hm_size_format($struct['rawsize']);
    array_shift($array);
    hm_assign_value($struct, $array, 'date');
    hm_assign_value($struct, $array, 'subject', true);
    $struct['from'] = hm_get_param_list($array, true);
    $struct['sender'] = hm_get_param_list($array, true);
    $struct['reply-to'] = hm_get_param_list($array, true);
    $struct['to'] = hm_get_param_list($array, true);
    $struct['cc'] = hm_get_param_list($array, true);
    $struct['bcc'] = hm_get_param_list($array, true);
    hm_assign_value($struct, $array, 'in-reply-to');
    hm_assign_value($struct, $array, 'message-id');
    array_shift($array);
    foreach (array('subject', 'from', 'date') as $val) {
        if (!isset($struct[$val])) {
            $struct[$val] = '';
        }
    }
    $struct['name'] = $struct['subject'];
    $struct['subs'] = hm_parse_top_level($array, $part.'.1');
    return $struct;
}

/**
 *  breaks up one section of the bodystructure by parens
 *
 *  @param          $array      array       whats left of the bodystructure response
 *  @return                     array       array with each section in an element
 *
 */
function hm_break_one_level($array) {
    $result = array();
    $pcount = 0;
    $i = 0;
    foreach ($array as $val) {
        if ($val == '(') {
            $pcount++;
        }
        $result[$i][] = $val;
        if ($val == ')') {
            $pcount--;
        }
        if ($pcount == 1) {
            $i++;
        }
    }
    return $result;
}

/**
 *  Assign a value to the $struct array and remove it from the $array array
 *
 *  @param          $struct     array       current structure array
 *  @param          $array      array       whats left of the bodystructure response
 *  @param          $name       string      field name to use as the array key
 *  @param          $decode     boolean     decode field value if true
 *  @param          $lower      boolean     lowercase value if true
 *  @param          $default    string      default to use in case value is not available
 *  @retrun                     viod        values modifed by ref instead (struct, array)
 *
 */
function hm_assign_value(&$struct, &$array, $name, $decode=false, $lower=false, $default=false) {
    $result = false;
    if (isset($array[0])) {
        $result = $array[0];
        array_shift($array);
    }
    if ($result) {
        if ($lower) {
            $result = strtolower($result);
        }
        if (!$decode) {
            $struct[$name] = $result;
        }
        else {
            $struct[$name] = hm_decode_field($result);
        }
    }
    elseif ($default) {
        $struct[$name] = $default;
    }
}

/**
 *  break up paren delimited parameter sections of an rfc822 part
 *
 *  @param          $array      array       whats left of the bodystructure response
 *  @param          $decode     boolean     decode value if true
 *  @return                     string      string value for the params
 *
 */
function hm_get_param_list(&$array, $decode=false) {
    $result = array();
    if (isset($array[0]) && strtolower($array[0]) == 'nil') {
        $result[] = 'NIL';
        array_shift($array);
    }
    elseif (isset($array[0])) {
        $end_paren = 0;
        $end = false;
        while (isset($array[0]) && !$end) {
            if ($array[0] == '(') {
                $end_paren++;
            }
            if ($array[0] == ')') {
                $end_paren--;
                if ($end_paren == 0) {
                    $end = true;
                }
            }
            if ($decode) {
                $result[] = hm_decode_field($array[0]);
            }
            else {
                $result[] = $array[0];
            }
            array_shift($array);
        }
        array_shift($result);
    }
    else {
        $result[] = 'NIL';
    }
    if (count($result > 1)) {
        $result = hm_parse_address_params($result);
    }
    return $result;    
}

/**
 *  helper function to hm_get_param_list. cleans up address values
 *
 *  @param          $result     array       broken up parameter values
 *  @return                     string      string value for the params
 *
 */
function hm_parse_address_params($result) {
    $string = '';
    $name = '';
    $mail = '';
    $domain = '';
    foreach ($result as $index => $val) {
        $test = ($index % 6);
        switch ($test) {
            case 1:
                $name = $val;
                break;
            case 3:
                $mail = $val;
                break;
            case 4:
                $domain = $val;
                break;
            case 5:
                if ($name && strtoupper($name) != 'NIL') {
                    $name = str_replace(array('"', "'"), '', $name);
                    if ($string != '') {
                        $string .= ', ';
                    }
                    if ($name != $mail.'@'.$domain) {
                        $string .= '"'.$name.'" ';
                    }
                    if ($mail && $domain) {
                        $string .= $mail.'@'.$domain;
                    }
                    $mail = '';
                    $name = '';
                    $domain = '';
                }
                break;
        }
    }
    return $string;
}

/**
 *  parse single part message part data
 *
 *  @param          $array      array       whats left of the bodystructure response
 *  @param          $part       string      message part number used to build filename
 *  @return                     array       multi-dimen bodystructure data
 *
 */ 
function hm_parse_single($array, $part) {
    if (empty($array)) {
        return array(); 
    }
    while ($array[0] == '(') {
        array_shift($array);
    }
    $params = array();
    $struct = array();
    hm_assign_value($struct, $array, 'type', false, true, 'unknown');
    hm_assign_value($struct, $array, 'subtype', false, true, 'unknown');
    if (substr($struct['subtype'], 0, 5) == 'plain') {
        $struct['subtype'] = 'plain';
    }
    if ($array[0] == '(') {
        array_shift($array);
        while ($array[0] != ')') {
            $params[] = $array[0];
            array_shift($array);
        }
        array_shift($array);
        if (!empty($params)) {
            $count = count($params);
            for ($i=0;$i<$count;$i+=2) {
                if (isset($params[$i]) && isset($params[$i + 1])) {
                    $struct[strtolower($params[$i])] = $params[$i + 1];
                }
            }
        }
    }
    else {
        array_shift($array);
    }
    $data = array('id' => false, 'description' => false,
                  'encoding' => true, 'rawsize' => false);
    foreach ($data as $key => $val) {
        if (isset($array[0])) {
            hm_assign_value($struct, $array, $key, false, $val);
        }
    }
    if (isset($struct['rawsize'])) {
        $struct['size'] = hm_size_format($struct['rawsize']);
    }
    else {
        $struct['size'] = 0;
        $struct['rawsize'] = '';
    }
    $count = count($array) - 1;
    for ($i=0;$i<$count;$i++) {
        if (strtolower($array[$i]) == 'filename') {
            $struct['name'] = $array[$i + 1];
            break;
        }
    }
    if (!isset($struct['name'])) {
        $base_name = str_replace('.', '_', $part);
        switch ($struct['type']) {
            case 'message':
            case 'text':
                switch ($struct['subtype']) {
                    case 'delivery-status':
                    case 'plain':
                        $struct['name'] = 'message_'.$base_name.'.txt';
                        break;
                    case 'html':
                        $struct['name'] = 'message_'.$base_name.'.html';
                        break;
                    case 'richtext':
                        $struct['name'] = 'message_'.$base_name.'.rtf';
                        break;
                    default:
                        $struct['name'] = 'message_'.$base_name.'.txt';
                        break;
                }
                break;
            case 'image':
                switch ($struct['subtype']) {
                    case 'gif':
                    case 'png':
                    case 'jpeg':
                    case 'jpg':
                    case 'pjpeg':
                        $struct['name'] = 'image_'.$base_name.'.'.$struct['subtype'];
                        break;
                    default:
                        $struct['name'] = '';
                        break;
                }
                break;
            default:
                $struct['name'] = '';
                break;
        }
    }
    return $struct;
}
?>
